//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_color.hpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#ifndef __LASERBOY_COLOR_DEFINITIONS__
#define __LASERBOY_COLOR_DEFINITIONS__

//############################################################################
class LaserBoy_color
{
public:
    LaserBoy_color( const u_char& _r = 0,
                    const u_char& _g = 0,
                    const u_char& _b = 0
                  )
            : r(_r), g(_g), b(_b)                      {}
    //------------------------------------------------------------------------
    LaserBoy_color(const LaserBoy_color& c)
            : r(c.r), g(c.g), b(c.b)                   {}
    //------------------------------------------------------------------------
    LaserBoy_color(const u_int& c)
            : r((c & 0x00ff0000) >> 16)
            , g((c & 0x0000ff00) >>  8)
            , b((c & 0x000000ff)      )                {}
    //------------------------------------------------------------------------
    LaserBoy_color& operator = (const LaserBoy_color& c)
                {
                    r = c.r;
                    g = c.g;
                    b = c.b;
                    return *this;
                }
    //------------------------------------------------------------------------
    LaserBoy_color& operator = (const u_char& c)
                {
                    r = c;
                    g = c;
                    b = c;
                    return *this;
                }
    //------------------------------------------------------------------------
    LaserBoy_color& operator = (const u_int& i)
                {
                    r = (i & 0x00ff0000) >> 16;
                    g = (i & 0x0000ff00) >>  8;
                    b = (i & 0x000000ff)      ;
                    return *this;
                }
    //------------------------------------------------------------------------
    bool operator == (const LaserBoy_color& c) const
            {
                if(    (r != c.r)
                    || (g != c.g)
                    || (b != c.b)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    bool operator != (const LaserBoy_color& c) const
            {
                if(    (r == c.r)
                    && (g == c.g)
                    && (b == c.b)
                  )
                    return false;
                return false;
            }
    //------------------------------------------------------------------------
    bool operator >  (const LaserBoy_color& c) const
            {
                return(   (  r << 16 |   g << 8 |   b)
                        > (c.r << 16 | c.g << 8 | c.b)
                      );
            }
    //------------------------------------------------------------------------
    bool operator <  (const LaserBoy_color& c) const
            {
                return(   (  r << 16 |   g << 8 |   b)
                        < (c.r << 16 | c.g << 8 | c.b)
                      );
            }
    //------------------------------------------------------------------------
    int  from_ifstream_ild (ifstream& in)
            {
                r = (u_char)in.get();
                g = (u_char)in.get();
                b = (u_char)in.get();
                return in.bad();
            }
    //------------------------------------------------------------------------
    void  to_ofstream_ild (ofstream& out) const
            {
                out.put((char)r);
                out.put((char)g);
                out.put((char)b);
                return;
            }
    //------------------------------------------------------------------------
    void  to_ofstream_txt (ofstream& out) const
            {
                out << setw(6) << (int)r << ' '
                    << setw(6) << (int)g << ' '
                    << setw(6) << (int)b
                    << endl;
                return;
            }
    //------------------------------------------------------------------------
    u_int as_uint()    {    return(r << 16 | g << 8 | b);    }
    //------------------------------------------------------------------------
    u_int intensity()  {    return(r + g + b) / 3;           }
    //------------------------------------------------------------------------
    u_char r, g, b;
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
