//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_vertex.hpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#ifndef __LASERBOY_3D_VERTEX_DEFINITIONS__
#define __LASERBOY_3D_VERTEX_DEFINITIONS__

//############################################################################
#include "LaserBoy_real_vertex.hpp"
#include "LaserBoy_wave.hpp"

//############################################################################
class LaserBoy_vertex : public LaserBoy_3D_short, public LaserBoy_color
{
public:
    LaserBoy_vertex(const short&   _x = 0,
                    const short&   _y = 0,
                    const short&   _z = 0,
                    const u_char&  _r = 255,
                    const u_char&  _g = 255,
                    const u_char&  _b = 255,
                    const u_char&  _k = LASERBOY_BLANKING_BIT | LASERBOY_UNIQUE_VERTEX,
                    const u_char&  _c = 55
                   )
                 : LaserBoy_3D_short (_x, _y, _z),
                   LaserBoy_color    (_r, _g, _b),
                   k                 (_k        ),
                   c                 (_c        )               {}
    //------------------------------------------------------------------------
    LaserBoy_vertex(const LaserBoy_vertex& p)
                 : LaserBoy_3D_short (p.x, p.y, p.z),
                   LaserBoy_color    (p.r, p.g, p.b),
                   k                 (p.k          ),
                   c                 (p.c          )            {}
    //------------------------------------------------------------------------
    LaserBoy_vertex(const LaserBoy_real_vertex& r) // cast to LaserBoy_vertex
                 : LaserBoy_3D_short ((short)round(r.x),
                                      (short)round(r.y),
                                      (short)round(r.z)
                                     ),
                   LaserBoy_color    (r.r, r.g, r.b),
                   k                 (r.k          ),
                   c                 (r.c          )            {}
    //------------------------------------------------------------------------
    LaserBoy_vertex(const LaserBoy_3D_short& s,
                    const LaserBoy_color&    rgb,
                    const u_char&            _k = LASERBOY_BLANKING_BIT | LASERBOY_UNIQUE_VERTEX,
                    const u_char&            _c = 0
                   )
                 : LaserBoy_3D_short (s  ),
                   LaserBoy_color    (rgb),
                   k                 (_k ),
                   c                 (_c )                      {}
    //------------------------------------------------------------------------
    LaserBoy_vertex(const LaserBoy_3D_short& s,
                    const u_char&            _c
                   )
                 : LaserBoy_3D_short (s                ),
                   LaserBoy_color    (255, 255, 255    ),
                   k                 (LASERBOY_BLANKING_BIT),
                   c                 (_c               )        {}
    //------------------------------------------------------------------------
    LaserBoy_vertex(const LaserBoy_3D_short& s)
                 : LaserBoy_3D_short (s                ),
                   LaserBoy_color    (255, 255, 255    ),
                   k                 (LASERBOY_BLANKING_BIT | LASERBOY_UNIQUE_VERTEX),
                   c                 (0                )        {}
    //------------------------------------------------------------------------
    LaserBoy_vertex(const LaserBoy_color& rgb)
                 : LaserBoy_3D_short (0, 0, 0          ),
                   LaserBoy_color    (rgb              ),
                   k                 (LASERBOY_BLANKING_BIT | LASERBOY_UNIQUE_VERTEX),
                   c                 (0                )        {}
    //------------------------------------------------------------------------
    LaserBoy_vertex& operator = (const LaserBoy_vertex& p)
                    {
                        x = p.x;
                        y = p.y;
                        z = p.z;
                        r = p.r;
                        g = p.g;
                        b = p.b;
                        k = p.k;
                        c = p.c;
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_vertex& operator = (const LaserBoy_3D_short& s) // only assign the coordinates
                    {
                        x = s.x;
                        y = s.y;
                        z = s.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_vertex  operator + (const LaserBoy_3D_short& s)
                    {
                        LaserBoy_vertex sum(*this);
                        sum.x += s.x;
                        sum.y += s.y;
                        sum.z += s.z;
                        return sum;
                    }
    //------------------------------------------------------------------------
    LaserBoy_vertex& operator += (const LaserBoy_3D_short& s)
                    {
                        x += s.x;
                        y += s.y;
                        z += s.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_vertex  operator - (const LaserBoy_3D_short& s)
                    {
                        LaserBoy_vertex diff(*this);
                        diff.x -= s.x;
                        diff.y -= s.y;
                        diff.z -= s.z;
                        return diff;
                    }
    //------------------------------------------------------------------------
    LaserBoy_vertex& operator -= (const LaserBoy_3D_short& s)
                    {
                        x -= s.x;
                        y -= s.y;
                        z -= s.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_vertex& operator *= (const LaserBoy_3D_short& s)
                    {
                        x = (short)(x * s.x);
                        y = (short)(y * s.y);
                        z = (short)(z * s.z);
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_vertex& operator *= (const LaserBoy_3D_double& f)
                    {
                        x = (short)(x * f.x);
                        y = (short)(y * f.y);
                        z = (short)(z * f.z);
                        return *this;
                    }
    //------------------------------------------------------------------------
    bool operator == (const LaserBoy_vertex& p) const
            {
                if(    (x != p.x)
                    || (y != p.y)
                    || (z != p.z)
                    || (r != p.r)
                    || (g != p.g)
                    || (b != p.b)
                    || ((k & LASERBOY_BLANKING_BIT) != (p.k & LASERBOY_BLANKING_BIT))
                    || (c != p.c)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    bool operator != (const LaserBoy_vertex& p) const
            {
                if(    (x == p.x)
                    && (y == p.y)
                    && (z == p.z)
                    && (r == p.r)
                    && (g == p.g)
                    && (b == p.b)
                    && ((k & LASERBOY_BLANKING_BIT) == (p.k & LASERBOY_BLANKING_BIT))
                    && (c == p.c)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    LaserBoy_vertex  operator - ()
                    {
                        LaserBoy_vertex negative;
                        negative.x = -x;
                        negative.y = -y;
                        negative.z = -z;
                        return negative;
                    }
    //------------------------------------------------------------------------
    bool is_equal_2D(const LaserBoy_vertex& p) const
            {
                if(    (x != p.x)
                    || (y != p.y)
                    || (r != p.r)
                    || (g != p.g)
                    || (b != p.b)
                    || ((k & LASERBOY_BLANKING_BIT) != (p.k & LASERBOY_BLANKING_BIT))
                    || (c != p.c)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    int from_ifstream_ild (ifstream& in, char format)
            {
                if(in.good())
                {
                    char p, q;
                    //--------------------------------------------------------
                    in.get(p);
                    in.get(q);
                    x = ((p & 0x00ff) << 8 | (q & 0x00ff));
                    //--------------------------------------------------------
                    in.get(p);
                    in.get(q);
                    y = ((p & 0x00ff) << 8 | (q & 0x00ff));
                    //--------------------------------------------------------
                    if(    format == LASERBOY_3D_FRAME
                        || format == LASERBOY_3D_FRAME_RGB
                      )
                    {
                        in.get(p);
                        in.get(q);
                        z = ((p & 0x00ff) << 8 | (q & 0x00ff));
                    }
                    //--------------------------------------------------------
                    k = (u_char)in.get();
                    if(    format == LASERBOY_3D_FRAME
                        || format == LASERBOY_2D_FRAME
                      )
                    {
                        c = (u_char)in.get();
                    }
                    else if(    format == LASERBOY_3D_FRAME_RGB
                             || format == LASERBOY_2D_FRAME_RGB
                           )
                    {
                        b = (u_char)in.get();
                        g = (u_char)in.get();
                        r = (u_char)in.get();
                    }
                    //--------------------------------------------------------
                    set_unique();
                }
                return in.bad();
            }
    //------------------------------------------------------------------------
    void to_ofstream_ild(ofstream& out, char format) const
            {
                switch(format)
                {
                    case LASERBOY_3D_FRAME:
                             out.put((char)((x & 0xff00) >> 8));
                             out.put((char) (x & 0x00ff)      );
                             out.put((char)((y & 0xff00) >> 8));
                             out.put((char) (y & 0x00ff)      );
                             out.put((char)((z & 0xff00) >> 8));
                             out.put((char) (z & 0x00ff)      );
                             out.put((char)k);
                             out.put((char)c);
                             break;
                    //--------------------------------------------------------
                    case LASERBOY_2D_FRAME:
                             out.put((char)((x & 0xff00) >> 8));
                             out.put((char) (x & 0x00ff)      );
                             out.put((char)((y & 0xff00) >> 8));
                             out.put((char) (y & 0x00ff)      );
                             out.put((char)k);
                             out.put((char)c);
                             break;
                    //--------------------------------------------------------
                    case LASERBOY_3D_FRAME_RGB:
                             out.put((char)((x & 0xff00) >> 8));
                             out.put((char) (x & 0x00ff)      );
                             out.put((char)((y & 0xff00) >> 8));
                             out.put((char) (y & 0x00ff)      );
                             out.put((char)((z & 0xff00) >> 8));
                             out.put((char) (z & 0x00ff)      );
                             out.put((char)k);
                             out.put((char)b);
                             out.put((char)g);
                             out.put((char)r);
                             break;
                    //--------------------------------------------------------
                    case LASERBOY_2D_FRAME_RGB:
                             out.put((char)((x & 0xff00) >> 8));
                             out.put((char) (x & 0x00ff)      );
                             out.put((char)((y & 0xff00) >> 8));
                             out.put((char) (y & 0x00ff)      );
                             out.put((char)k);
                             out.put((char)b);
                             out.put((char)g);
                             out.put((char)r);
                             break;
                    //--------------------------------------------------------
                }
                return;
            }
    //------------------------------------------------------------------------
    void select       ()       { k |=  LASERBOY_SELECTED_VERTEX;  return;    }
    void deselect     ()       { k &= ~LASERBOY_SELECTED_VERTEX;  return;    }
    void toggle_select()       { is_selected() ? deselect() : select();      }
    bool is_selected  () const { return (bool)(k & LASERBOY_SELECTED_VERTEX);}
    //------------------------------------------------------------------------
    LaserBoy_vertex& blank  () { k |=  LASERBOY_BLANKING_BIT;  return *this; }
    LaserBoy_vertex& unblank() { k &= ~LASERBOY_BLANKING_BIT;  return *this; }
    //------------------------------------------------------------------------
    bool is_blank     () const { return (bool)(k & LASERBOY_BLANKING_BIT);   }
    bool is_lit       () const { return !is_blank();                         }
    //------------------------------------------------------------------------
    void set_unique   ()       { k |=  LASERBOY_UNIQUE_VERTEX;  return;      }
    void clear_unique ()       { k &= ~LASERBOY_UNIQUE_VERTEX;  return;      }
    bool is_unique    () const { return (bool)(k & LASERBOY_UNIQUE_VERTEX);  }
    //------------------------------------------------------------------------
    bool is_black     () const
                    {
                        return (    is_lit()
                                 && r == 0
                                 && g == 0
                                 && b == 0
                               );
                    }
    //------------------------------------------------------------------------
    bool  is_color () const
            {
                return (    is_lit()
                         && (r + g + b)
                       );
            }
    //------------------------------------------------------------------------
    void  clear()
            {
                x = 0;
                y = 0;
                z = 0;
                r = 0;
                g = 0;
                b = 0;
                k = LASERBOY_BLANKING_BIT | LASERBOY_UNIQUE_VERTEX;
                c = 0;
            }
    //------------------------------------------------------------------------
    LaserBoy_3D_short position() const {  return (LaserBoy_3D_short)(*this); }
    //------------------------------------------------------------------------
    int color() const
        {
            return   is_lit()
                   ? (int)(   r << 16
                            | g <<  8
                            | b
                          )
                   : -1; // is blank
        }
    //------------------------------------------------------------------------
    LaserBoy_color as_LaserBoy_color() const
        {
            return (LaserBoy_color)(*this);
        }
    //------------------------------------------------------------------------
    LaserBoy_3D_short bit_masked_position(const int signal_bit_mask[8]) const
    {
        LaserBoy_3D_short masked = *this;
        masked.x &= short_bit_mask[signal_bit_mask[0]];
        masked.y &= short_bit_mask[signal_bit_mask[1]];
        masked.z &= short_bit_mask[signal_bit_mask[5]];
        return masked;
    }
    //------------------------------------------------------------------------
    LaserBoy_color bit_masked_color(const int signal_bit_mask[8]) const
    {
        LaserBoy_color masked = *this;
        masked.r &= (short_bit_mask[signal_bit_mask[2]] >> 7);
        masked.g &= (short_bit_mask[signal_bit_mask[3]] >> 7);
        masked.b &= (short_bit_mask[signal_bit_mask[4]] >> 7);
        return masked;
    }
    //------------------------------------------------------------------------
    void to_ofstream_wave(ofstream&             out,
                          LaserBoy_wave_header& header,
                          const int             signal_bit_mask[8],
                          const bool&           end_of_frame,
                          const bool&           unique_frame
                         ) const
    {
        short  xx, yy, zz, color;
        u_char rr, gg, bb;
        //--------------------------------------------------------------------
        xx =  x &  short_bit_mask[signal_bit_mask[0]];
        yy =  y &  short_bit_mask[signal_bit_mask[1]];
        zz =  z &  short_bit_mask[signal_bit_mask[5]];
        //--------------------------------------------------------------------
        if(is_blank())
            rr = gg = bb = 0;
        else
        {
            rr = r & (short_bit_mask[signal_bit_mask[2]] >> 7);
            gg = g & (short_bit_mask[signal_bit_mask[3]] >> 7);
            bb = b & (short_bit_mask[signal_bit_mask[4]] >> 7);
        }
        //--------------------------------------------------------------------
        out.put(    xx & 0x00ff      ); // x
        out.put((   xx & 0xff00) >> 8);
        //--------------------------------------------------------------------
        out.put(    yy & 0x00ff      ); // y
        out.put((   yy & 0xff00) >> 8);
        //--------------------------------------------------------------------
        color = (rr << 7) + end_of_frame;
        out.put( color & 0x00ff      ); // red
        out.put((color & 0xff00) >> 8);
        //--------------------------------------------------------------------
        color = (gg << 7) + unique_frame;
        out.put( color & 0x00ff      ); // green
        out.put((color & 0xff00) >> 8);
        //--------------------------------------------------------------------
        color = (bb << 7) + (bool)(k & LASERBOY_UNIQUE_VERTEX);
        out.put( color & 0x00ff      ); // blue
        out.put((color & 0xff00) >> 8);
        //--------------------------------------------------------------------
        switch(abs(header.signal_id[5]))
        {
            case LASERBOY_SIGNAL_NO_SIGNAL:
                    out.put(0x00);
                    out.put(0x00);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_Z_POSITION:
                    out.put(    zz & 0x00ff      ); // z
                    out.put((   zz & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_MONO_TTL:
                    color = (is_lit()) ? (LASERBOY_MAX_COLOR_SHORT) : (0);
                    out.put( color & 0x00ff      );
                    out.put((color & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_MONO_OR_ANALOG:
                    color =  ((rr | gg | bb) << 7);
                    out.put( color & 0x00ff      );
                    out.put((color & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_MONO_WEIGHTED_ANALOG:
                    color = (((76 * rr) + (150 * gg) + (28 * bb)) / 254) << 7;
                    out.put( color & 0x00ff      );
                    out.put((color & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_MONO_AVG_ANALOG:
                    color = ((rr + gg + bb) / 3) << 7;
                    out.put( color & 0x00ff      );
                    out.put((color & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
        }
        header.num_samples++;
        return;
    }
    //------------------------------------------------------------------------
    void to_ofstream_wave_inverted(ofstream&              out,
                                   LaserBoy_wave_header&  header,
                                   const int              signal_bit_mask[8],
                                   const bool&            end_of_frame,
                                   const bool&            unique_frame
                                  ) const
    {
        short  xx, yy, zz, color;
        u_char rr, gg, bb;
        //--------------------------------------------------------------------
        xx =  x &  short_bit_mask[signal_bit_mask[0]];
        yy =  y &  short_bit_mask[signal_bit_mask[1]];
        zz =  z &  short_bit_mask[signal_bit_mask[5]];
        //--------------------------------------------------------------------
        if(is_blank())
            rr = gg = bb = 0;
        else
        {
            rr = r & (short_bit_mask[signal_bit_mask[2]] >> 7);
            gg = g & (short_bit_mask[signal_bit_mask[3]] >> 7);
            bb = b & (short_bit_mask[signal_bit_mask[4]] >> 7);
        }
        //--------------------------------------------------------------------
        out.put(   -xx & 0x00ff      ); // x
        out.put((  -xx & 0xff00) >> 8);
        //--------------------------------------------------------------------
        out.put(   -yy & 0x00ff      ); // y
        out.put((  -yy & 0xff00) >> 8);
        //--------------------------------------------------------------------
        color = (rr << 7) + end_of_frame;
        color = -color;
        out.put( color & 0x00ff      ); // red
        out.put((color & 0xff00) >> 8);
        //--------------------------------------------------------------------
        color = (gg << 7) + unique_frame;
        color = -color;
        out.put( color & 0x00ff      ); // green
        out.put((color & 0xff00) >> 8);
        //--------------------------------------------------------------------
        color = (bb << 7) + (bool)(k & LASERBOY_UNIQUE_VERTEX);
        color = -color;
        out.put( color & 0x00ff      ); // blue
        out.put((color & 0xff00) >> 8);
        //--------------------------------------------------------------------
        switch(abs(header.signal_id[5]))
        {
            case LASERBOY_SIGNAL_NO_SIGNAL:
                    out.put(0x00);
                    out.put(0x00);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_Z_POSITION:
                    out.put(   -zz & 0x00ff      ); // z
                    out.put((  -zz & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_MONO_TTL:
                    color = (is_lit()) ? (-LASERBOY_MAX_COLOR_SHORT) : (0);
                    out.put( color & 0x00ff      );
                    out.put((color & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_MONO_OR_ANALOG:
                    color =  ((rr | gg | bb) << 7);
                    color = -color;
                    out.put( color & 0x00ff      );
                    out.put((color & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_MONO_WEIGHTED_ANALOG:
                    color = (((76 * rr) + (150 * gg) + (28 * bb)) / 254) << 7;
                    color = -color;
                    out.put( color & 0x00ff      );
                    out.put((color & 0xff00) >> 8);
                    break;
            //----------------------------------------------------------------
            case LASERBOY_SIGNAL_MONO_AVG_ANALOG:
                    color = ((rr + gg + bb) / 3) << 7;
                    color = -color;
                    out.put( color & 0x00ff      );
                    out.put((color & 0xff00) >> 8);
                    break;
        //--------------------------------------------------------------------
        }
        header.num_samples++;
        return;
    }
    //------------------------------------------------------------------------
    u_char  k, c;
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
